/*
	RT Bench - Benchmarks various hash functions for Rainbow Tables
	Copyright (C) 2012 Steve Thomas <steve AT tobtu DOT com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/

#include "keyspace.h"
#include <string.h>
#include <stdlib.h>
#include <vector>
#include <fstream>
#include <sstream>
#include <string>
#include <map>
#include <list>

using namespace std;

void readLineTrim(ifstream &fin, string &line)
{
	uint32_t lastNonBlank = 0xffffffff;
	char ch;

	line = "";
	ch = fin.get();
	while (fin.good())
	{
		if (ch == '\r')
		{
			if (fin.peek() == '\n')
			{
				fin.get();
			}
			break;
		}
		if (ch == '\n')
		{
			break;
		}
		if (ch == ' ' || ch == '\t')
		{
			if (lastNonBlank != 0xffffffff)
			{
				line += ch;
			}
		}
		else
		{
			line += ch;
			lastNonBlank = line.length();
		}
		ch = fin.get();
	}
	if (lastNonBlank != 0xffffffff && lastNonBlank != line.length())
	{
		line = line.substr(0, lastNonBlank);
	}
}

string hex2bin(const string &hexStr)
{
	string ret;
	uint32_t a, len = hexStr.length();
	char ch, ch2;

	if ((len & 1) != 0)
	{
		return "";
	}
	for (a = 0; a < len; a += 2)
	{
		ch = hexStr[a];
		if (ch >= '0' && ch <= '9')
		{
			ch2 = (ch - '0') << 4;
		}
		else if (ch >= 'A' && ch <= 'F')
		{
			ch2 = (ch - 'A' + 10) << 4;
		}
		else if (ch >= 'a' && ch <= 'f')
		{
			ch2 = (ch - 'a' + 10) << 4;
		}
		else
		{
			return "";
		}
		ch = hexStr[a + 1];
		if (ch >= '0' && ch <= '9')
		{
			ch2 |= ch - '0';
		}
		else if (ch >= 'A' && ch <= 'F')
		{
			ch2 |= ch - 'A' + 10;
		}
		else if (ch >= 'a' && ch <= 'f')
		{
			ch2 |= ch - 'a' + 10;
		}
		else
		{
			return "";
		}
		ret += ch2;
	}
	return ret;
}

KeySpace::KeySpace(const KeySpace &ks)
{
	copyObject(ks);
}

KeySpace::KeySpace(const char *fileName, uint32_t charaterEncoding)
{
	ifstream fin(fileName, ifstream::in);
	string line, name, charSet;
	map<string,string> charSets;
	map<string,string>::iterator charSetMapIt;
	list<string> subKeySpace;
	list<list<string> > subKeySpaces;
	list<string>::iterator charSetIt;
	list<list<string> >::iterator subKeySpaceIt;
	size_t pos, pos2;
	uint32_t len, lineNumber = 0;
	bool multiCharNames = false;
	char ch[2] = {0, 0};

	if (!fin.good())
	{
		reportError("Can't open key space file '%s'\n", fileName);
		exit(1);
	}

	// Read character sets
	while (1)
	{
		// Read line
		readLineTrim(fin, line);
		lineNumber++;
		if (!fin.good() && !fin.eof())
		{
			reportError("Unable to read line #%u from key space file '%s'\n", lineNumber, fileName);
			exit(1);
		}
		if (line.length() == 0 || line[0] == '#')
		{
			if (fin.eof())
			{
				reportError("'%s' is not a valid key space file\n", fileName);
				exit(1);
			}
			continue;
		}
		if (line[line.length() - 1] != ']')
		{
			break;
		}

		// Get character set name
		pos = line.find_first_of(" \t", 0, 2);
		if (pos == string::npos)
		{
			break;
		}
		name = line.substr(0, pos);

		// Get character set
		pos = line.find_first_not_of(" \t", pos, 2);
		if (pos == string::npos)
		{
			break; // this is not even possible... meh
		}
		if (line[pos] == '[')
		{
			charSet = line.substr(pos + 1, line.length() - pos - 2);
		}
		else if (line[pos] == 'h' && line[pos + 1] == '[')
		{
			charSet = hex2bin(line.substr(pos + 2, line.length() - pos - 3));
			if (charSet.length() == 0 && line.length() - pos - 3 != 0)
			{
				reportError("'%s' has an invalid character set on line #%u\n", fileName, lineNumber);
				exit(1);
			}
		}
		else
		{
			break;
		}
		if (charSet.length() == 0)
		{
			reportError("'%s' has an empty character set on line #%u\n", fileName, lineNumber);
			exit(1);
		}

		// Insert into map
		if (charSets.find(name) == charSets.end())
		{
			if (name.length() != 1)
			{
				multiCharNames = true;
			}
			charSets[name] = charSet;
		}
		else
		{
			reportError("'%s' has a duplicate character set name on line #%u\n", fileName, lineNumber);
			exit(1);
		}
	}
	if (line[line.length() - 1] == ']')
	{
		reportError("line #%u, '%s' is not a valid key space file\n", lineNumber, fileName);
		exit(1);
	}

	// Read sub key spaces
	do
	{
		if (multiCharNames)
		{
			pos = 0;
			while (pos != string::npos)
			{
				// Get character set name
				pos2 = line.find_first_of(" \t", pos, 2);
				if (pos2 == string::npos)
				{
					pos2 = line.length();
				}
				name = line.substr(pos, pos2 - pos);

				// Get character set
				charSetMapIt = charSets.find(name);
				if (charSetMapIt == charSets.end())
				{
					reportError("Character set name '%s' not found on line #%u from key space file '%s'\n", name.c_str(), lineNumber, fileName);
					exit(1);
				}
				subKeySpace.push_back(charSetMapIt->second);

				if (pos2 == line.length())
				{
					break;
				}
				pos = line.find_first_not_of(" \t", pos2, 2);
			}
		}
		else
		{
			len = line.length();
			for (pos = 0; pos < len; pos++)
			{
				*ch = line[pos];
				if (*ch != ' ' && *ch != '\t')
				{
					charSetMapIt = charSets.find(ch);
					if (charSetMapIt == charSets.end())
					{
						reportError("Character set name '%c' not found on line #%u from key space file '%s'\n", *ch, lineNumber, fileName);
						exit(1);
					}
					subKeySpace.push_back(charSetMapIt->second);
				}
			}
		}
		subKeySpaces.push_back(subKeySpace);
		subKeySpace.clear();

		// Read line
		do
		{
			readLineTrim(fin, line);
			lineNumber++;
			if (!fin.good() && !fin.eof())
			{
				reportError("Unable to read line #%u from key space file '%s'\n", lineNumber, fileName);
				exit(1);
			}
		} while ((line.length() == 0 || line[0] == '#') && fin.good());
	} while (line.length() != 0);

	init(subKeySpaces, charaterEncoding);
}

KeySpace::KeySpace(list<list<string> > subKeySpaces, uint32_t charaterEncoding)
{
	init(subKeySpaces, charaterEncoding);
}

void KeySpace::init(list<list<string> > subKeySpaces, uint32_t charaterEncoding)
{
	uint64_t keySpace, keySpaceBoundary = 0;
	list<string> uniqueCharSets;
	list<string>::iterator uniqueCharSetsIt;
	list<string>::reverse_iterator charSetItRev;
	list<string>::iterator charSetIt;
	list<list<string> >::iterator subKeySpaceIt;
	uint32_t a, b, charSetLen = 0, maxPwLen = 1, minPwLen = 0xffffffff;
	int32_t charSetDiffLen;

	m_curSubKeySpace = 0;
	m_numSubKeySpaces = subKeySpaces.size();
	m_charaterEncoding = charaterEncoding;
	m_offset = 0;
	if ((charaterEncoding & CE_UTF16_BE) != 0)
	{
		m_offset = 1;
	}
	else if ((charaterEncoding & CE_UTF32_BE) != 0)
	{
		m_offset = 3;
	}
	m_keySpaceBoundaries = new uint64_t [m_numSubKeySpaces + 1];
	m_passwordLengths    = new uint32_t [m_numSubKeySpaces];
	m_divShortCut        = new uint32_t [m_numSubKeySpaces];
	m_numShortCutChars   = new int      [m_numSubKeySpaces];
	m_charSetLengths     = new uint32_t*[m_numSubKeySpaces];
	m_subKeySpaces       = new char**   [m_numSubKeySpaces];

	m_keySpaceBoundaries[0] = 0;
	for (a = 0, subKeySpaceIt = subKeySpaces.begin(); subKeySpaceIt != subKeySpaces.end(); ++subKeySpaceIt, a++)
	{
		if (maxPwLen < subKeySpaceIt->size())
		{
			maxPwLen = subKeySpaceIt->size();
		}
		if (minPwLen > subKeySpaceIt->size())
		{
			minPwLen = subKeySpaceIt->size();
		}
		keySpace = 1;
		m_passwordLengths[a] = subKeySpaceIt->size();
		m_charSetLengths[a] = new uint32_t[subKeySpaceIt->size()];
		for (b = subKeySpaceIt->size() - 1, charSetItRev = subKeySpaceIt->rbegin(); charSetItRev != subKeySpaceIt->rend(); ++charSetItRev, b--)
		{
			// Calculate key space
			keySpace *= charSetItRev->length();
			if (keySpace < UINT64_C(0x100000000))
			{
				m_divShortCut[a] = (uint32_t) keySpace;
				m_numShortCutChars[a] = subKeySpaceIt->size() - b;
			}

			// Find unique character sets
			m_charSetLengths[a][b] = charSetItRev->length();
			for (uniqueCharSetsIt = uniqueCharSets.begin(); uniqueCharSetsIt != uniqueCharSets.end(); ++uniqueCharSetsIt)
			{
				charSetDiffLen = m_charSetLengths[a][b] - uniqueCharSetsIt->length();
				if (charSetDiffLen > 0)
				{
					if (charSetItRev->find(*uniqueCharSetsIt) != string::npos)
					{
						*uniqueCharSetsIt = *charSetItRev;
						charSetLen += charSetDiffLen;
						break;
					}
				}
				else if (charSetDiffLen < 0)
				{
					if (uniqueCharSetsIt->find(*charSetItRev) != string::npos)
					{
						break;
					}
				}
				else if (*uniqueCharSetsIt == *charSetItRev)
				{
					break;
				}
			}
			if (uniqueCharSetsIt == uniqueCharSets.end())
			{
				uniqueCharSets.push_back(charSetItRev->c_str());
				charSetLen += charSetItRev->length();
			}
		}
		if (keySpace == m_divShortCut[a])
		{
			m_numShortCutChars[a] = 0;
		}
		else if (keySpace / m_divShortCut[a] >= UINT64_C(0x100000000))
		{
			m_numShortCutChars[a] = -m_numShortCutChars[a];
		}
		keySpaceBoundary += keySpace;
		m_keySpaceBoundaries[a + 1] = keySpaceBoundary;
	}
	m_keySpace = keySpaceBoundary;
	m_maxPwLen = maxPwLen;
	m_minPwLen = minPwLen;
	m_charSetPoses = new uint32_t[maxPwLen];
	m_charSetPosesTemp = new uint32_t[maxPwLen];
	for (a = 0; a < maxPwLen; a++)
	{
		m_charSetPoses[0] = 0;
	}
	if ((m_charaterEncoding & CE_80_NULL_PADDED) != 0)
	{
		maxPwLen++;
	}
	if ((m_charaterEncoding & (CE_UTF16_LE | CE_UTF16_BE)) != 0)
	{
		maxPwLen *= 2;
	}
	else if ((m_charaterEncoding & (CE_UTF32_LE | CE_UTF32_BE)) != 0)
	{
		maxPwLen *= 4;
	}
	else if ((m_charaterEncoding & CE_ASCII_BIN8) == 0)
	{
		reportError("Charater encoding not supported\n");
		exit(1);
	}
	maxPwLen = (maxPwLen + 3) / 4;
	m_pwBlocks = maxPwLen;
	m_pw = new uint32_t[maxPwLen + 1];
	m_pwTemp = new uint32_t[maxPwLen + 1];

	m_charSet = new char[charSetLen + 1];
	m_charSet[charSetLen] = 0;
	a = 0;
	for (uniqueCharSetsIt = uniqueCharSets.begin(); uniqueCharSetsIt != uniqueCharSets.end(); ++uniqueCharSetsIt)
	{
		charSetLen = uniqueCharSetsIt->length();
		memcpy(m_charSet + a, uniqueCharSetsIt->c_str(), charSetLen);
		a += charSetLen;
	}
	uniqueCharSets.clear();
	for (a = 0, subKeySpaceIt = subKeySpaces.begin(); subKeySpaceIt != subKeySpaces.end(); ++subKeySpaceIt, a++)
	{
		m_subKeySpaces[a] = new char*[subKeySpaceIt->size()];
		for (b = 0, charSetIt = subKeySpaceIt->begin(); charSetIt != subKeySpaceIt->end(); ++charSetIt, b++)
		{
			m_subKeySpaces[a][b] = strstr(m_charSet, charSetIt->c_str());
		}
	}

	reset(0);
}

KeySpace::~KeySpace()
{
	cleanUp();
}

KeySpace& KeySpace::operator=(const KeySpace &ks)
{
	if (this != &ks)
	{
		cleanUp();
		copyObject(ks);
	}
	return *this;
}

void KeySpace::cleanUp()
{
	uint32_t a;

	for (a = 0; a < m_numSubKeySpaces; a++)
	{
		delete [] m_subKeySpaces[a];
		delete [] m_charSetLengths[a];
	}
	delete [] m_keySpaceBoundaries;
	delete [] m_passwordLengths;
	delete [] m_divShortCut;
	delete [] m_numShortCutChars;
	delete [] m_charSetPoses;
	delete [] m_charSetPosesTemp;
	delete [] m_charSetLengths;
	delete [] m_charSet;
	delete [] m_subKeySpaces;
	delete [] m_pw;
	delete [] m_pwTemp;
}

void KeySpace::copyObject(const KeySpace &ks)
{
	uint32_t a, b, tmp, charSetLen = 1;

	m_keySpace = ks.m_keySpace;
	m_numSubKeySpaces = ks.m_numSubKeySpaces;
	m_curSubKeySpace = ks.m_curSubKeySpace;
	m_maxPwLen = ks.m_maxPwLen;
	m_minPwLen = ks.m_minPwLen;
	m_charaterEncoding = ks.m_charaterEncoding;
	m_offset = ks.m_offset;
	m_pwBlocks = ks.m_pwBlocks;

	m_keySpaceBoundaries  = new uint64_t [m_numSubKeySpaces + 1];
	m_passwordLengths     = new uint32_t [m_numSubKeySpaces];
	m_divShortCut         = new uint32_t [m_numSubKeySpaces];
	m_numShortCutChars    = new int      [m_numSubKeySpaces];
	m_charSetLengths      = new uint32_t*[m_numSubKeySpaces];
	m_subKeySpaces        = new char**   [m_numSubKeySpaces];
	m_charSetPoses        = new uint32_t [m_maxPwLen];
	m_charSetPosesTemp    = new uint32_t [m_maxPwLen];
	m_pw                  = new uint32_t [m_pwBlocks + 1];
	m_pwTemp              = new uint32_t [m_pwBlocks + 1];
	memcpy(m_keySpaceBoundaries,  ks.m_keySpaceBoundaries,  sizeof(uint64_t) * (m_numSubKeySpaces + 1));
	memcpy(m_passwordLengths,     ks.m_passwordLengths,     sizeof(uint32_t) * m_numSubKeySpaces);
	memcpy(m_divShortCut,         ks.m_divShortCut,         sizeof(uint32_t) * m_numSubKeySpaces);
	memcpy(m_numShortCutChars,    ks.m_numShortCutChars,    sizeof(uint32_t) * m_numSubKeySpaces);
	memcpy(m_charSetPoses,        ks.m_charSetPoses,        sizeof(uint32_t) * m_maxPwLen);
	memcpy(m_pw,                  ks.m_pw,                  sizeof(uint32_t) * (m_pwBlocks + 1));
	for (a = 0; a < m_numSubKeySpaces; a++)
	{
		m_charSetLengths[a] = new uint32_t[m_passwordLengths[a]];
		memcpy(m_charSetLengths[a], ks.m_charSetLengths[a], sizeof(uint32_t) * m_passwordLengths[a]);
		for (b = 0; b < m_passwordLengths[a]; b++)
		{
			tmp = m_charSetLengths[a][b] + (ks.m_subKeySpaces[a][b] - ks.m_charSet);
			if (charSetLen < tmp)
			{
				charSetLen = tmp;
			}
		}
	}
	m_charSet = new char[charSetLen];
	memcpy(m_charSet, ks.m_charSet, sizeof(char) * charSetLen);
	for (a = 0; a < m_numSubKeySpaces; a++)
	{
		m_subKeySpaces[a] = new char*[m_passwordLengths[a]];
		for (b = 0; b < m_passwordLengths[a]; b++)
		{
			m_subKeySpaces[a][b] = m_charSet + (ks.m_subKeySpaces[a][b] - ks.m_charSet);
		}
	}
}

bool KeySpace::reset(uint64_t pwNum)
{
	return reset(pwNum, m_pw, m_charSetPoses, m_curSubKeySpace);
}

const bool KeySpace::reset(uint64_t pwNum, uint32_t *pwOut, uint32_t *charSetPoses, uint32_t &curSubKeySpace)
{
	uint32_t *charSetLengths, *pw;
	char **subKeySpaces;
	uint32_t curSubKeySpaceTemp, charSetLength, divisor, remainder, quotient, offset;
	int a, t, passwordLength, numShortCutChars;

	if (pwNum >= m_keySpace)
	{
		return true;
	}
	pw = pwOut + 1;
	offset = m_offset;
	curSubKeySpaceTemp = 0;
	while (pwNum >= m_keySpaceBoundaries[curSubKeySpaceTemp + 1])
	{
		curSubKeySpaceTemp++;
	}
	pwNum           -= m_keySpaceBoundaries[curSubKeySpaceTemp];
	passwordLength   = (int) m_passwordLengths[curSubKeySpaceTemp];
	numShortCutChars = (int) m_numShortCutChars[curSubKeySpaceTemp];
	charSetLengths   = m_charSetLengths[curSubKeySpaceTemp];
	subKeySpaces     = m_subKeySpaces[curSubKeySpaceTemp];
	curSubKeySpace   = curSubKeySpaceTemp;
	t = (int) m_pwBlocks;
	for (a = passwordLength >> 2; a < t; a++)
	{
		pw[a] = 0;
	}
	if (numShortCutChars > 0)
	{
		divisor = m_divShortCut[curSubKeySpaceTemp];
		DIV_MOD_64_RET_32_32(pwNum, divisor, remainder, quotient);

		numShortCutChars = passwordLength - numShortCutChars;
		if ((m_charaterEncoding & CE_ASCII_BIN8) != 0)
		{
			// 8 bit
			for (a = passwordLength - 1; a >= numShortCutChars; a--)
			{
				charSetLength = charSetLengths[a];
				charSetPoses[a] = t = remainder % charSetLength;
				remainder /= charSetLength;
				((char*)pw)[a] = subKeySpaces[a][t];
			}
			for (; a >= 0; a--)
			{
				charSetLength = charSetLengths[a];
				charSetPoses[a] = t = quotient % charSetLength;
				quotient /= charSetLength;
				((char*)pw)[a] = subKeySpaces[a][t];
			}
		}
		else if ((m_charaterEncoding & (CE_UTF16_LE | CE_UTF16_BE)) != 0)
		{
			// 16 bit
			for (a = passwordLength - 1; a >= numShortCutChars; a--)
			{
				charSetLength = charSetLengths[a];
				charSetPoses[a] = t = remainder % charSetLength;
				remainder /= charSetLength;
				((char*)pw)[2 * a + offset] = subKeySpaces[a][t];
			}
			for (; a >= 0; a--)
			{
				charSetLength = charSetLengths[a];
				charSetPoses[a] = t = quotient % charSetLength;
				quotient /= charSetLength;
				((char*)pw)[2 * a + offset] = subKeySpaces[a][t];
			}
		}
		else // if ((m_charaterEncoding & (CE_UTF32_LE | CE_UTF32_BE)) != 0)
		{
			// 32 bit
			for (a = passwordLength - 1; a >= numShortCutChars; a--)
			{
				charSetLength = charSetLengths[a];
				charSetPoses[a] = t = remainder % charSetLength;
				remainder /= charSetLength;
				((char*)pw)[4 * a + offset] = subKeySpaces[a][t];
			}
			for (; a >= 0; a--)
			{
				charSetLength = charSetLengths[a];
				charSetPoses[a] = t = quotient % charSetLength;
				quotient /= charSetLength;
				((char*)pw)[4 * a + offset] = subKeySpaces[a][t];
			}
		}
	}
	else if (passwordLength > 1)
	{
		quotient = (uint32_t) pwNum;
		if ((m_charaterEncoding & CE_ASCII_BIN8) != 0)
		{
			// 8 bit
			for (a = passwordLength - 1; a >= 0; a--)
			{
				charSetLength = charSetLengths[a];
				charSetPoses[a] = t = quotient % charSetLength;
				quotient /= charSetLength;
				((char*)pw)[a] = subKeySpaces[a][t];
			}
		}
		else if ((m_charaterEncoding & (CE_UTF16_LE | CE_UTF16_BE)) != 0)
		{
			// 16 bit
			for (a = passwordLength - 1; a >= 0; a--)
			{
				charSetLength = charSetLengths[a];
				charSetPoses[a] = t = quotient % charSetLength;
				quotient /= charSetLength;
				((char*)pw)[2 * a + offset] = subKeySpaces[a][t];
			}
		}
		else // if ((m_charaterEncoding & (CE_UTF32_LE | CE_UTF32_BE)) != 0)
		{
			// 32 bit
			for (a = passwordLength - 1; a >= 0; a--)
			{
				charSetLength = charSetLengths[a];
				charSetPoses[a] = t = quotient % charSetLength;
				quotient /= charSetLength;
				((char*)pw)[4 * a + offset] = subKeySpaces[a][t];
			}
		}
	}
	else
	{
		charSetPoses[0] = (uint32_t) pwNum;
		((char*)pw)[offset] = subKeySpaces[0][pwNum];
	}
	pwOut[0] = passwordLength;
	if ((m_charaterEncoding & CE_80_NULL_PADDED) != 0)
	{
		if (m_charaterEncoding == (CE_ASCII_BIN8 | CE_80_NULL_PADDED))
		{
			((char*)pw)[passwordLength] = (char) 0x80;
		}
		else if ((m_charaterEncoding & (CE_UTF16_LE | CE_UTF16_BE)) != 0)
		{
			((char*)pw)[2 * passwordLength + offset] = (char) 0x80;
		}
		else // if ((m_charaterEncoding & (CE_UTF32_LE | CE_UTF32_BE)) != 0)
		{
			((char*)pw)[4 * passwordLength + offset] = (char) 0x80;
		}
	}
	return false;
}

uint32_t KeySpace::next(vector4 *pwOut, uint32_t numPws, uint32_t vecSize)
{
	char *charSet;
	uint32_t *pw;
	uint32_t a, b, c, curSubKeySpace, passwordLength, charSetPos, charSetLength, offset, pwBlocks;

	pw             = m_pw;
	offset         = m_offset;
	pwBlocks       = m_pwBlocks;
	curSubKeySpace = m_curSubKeySpace;
	passwordLength = m_passwordLengths[curSubKeySpace];
	charSet        = m_subKeySpaces[curSubKeySpace][passwordLength - 1];
	charSetPos     = m_charSetPoses[passwordLength - 1];
	charSetLength  = m_charSetLengths[curSubKeySpace][passwordLength - 1];
	for (a = 0; a < numPws; a++)
	{
		if ((m_charaterEncoding & CE_ASCII_BIN8) != 0)
		{
			// 8 bit
			((char*)(pw+1))[passwordLength - 1] = charSet[charSetPos];
		}
		else if ((m_charaterEncoding & (CE_UTF16_LE | CE_UTF16_BE)) != 0)
		{
			// 16 bit
			((char*)(pw+1))[2 * (passwordLength - 1) + offset] = charSet[charSetPos];
		}
		else // if ((m_charaterEncoding & (CE_UTF32_LE | CE_UTF32_BE)) != 0)
		{
			// 32 bit
			((char*)(pw+1))[4 * (passwordLength - 1) + offset] = charSet[charSetPos];
		}
		for (b = 0, c = 0; b < pwBlocks; b++, c += vecSize)
		{
			pwOut[c].data[a] = pw[b];
		}
		charSetPos++;
		if (charSetPos >= charSetLength)
		{
			if (cycleLastChar())
			{
				reset(0);
			}
			charSetPos = 0;
			curSubKeySpace = m_curSubKeySpace;
			passwordLength = m_passwordLengths[curSubKeySpace];
			charSet        = m_subKeySpaces[curSubKeySpace][passwordLength - 1];
			charSetLength  = m_charSetLengths[curSubKeySpace][passwordLength - 1];
		}
	}
	m_charSetPoses[passwordLength - 1] = charSetPos;
	return a;
}

uint32_t KeySpace::get(vector4 *pwOut, uint32_t numPws, uint32_t vecSize, uint64_t *pwNums)
{
	uint32_t *pw, *charSetPoses;
	uint32_t a, b, c, curSubKeySpace, pwBlocks;

	pwBlocks       = m_pwBlocks;
	pw             = m_pwTemp;
	charSetPoses   = m_charSetPosesTemp; // Thrown away
	curSubKeySpace = m_curSubKeySpace;
	for (a = 0; a < numPws; a++)
	{
		if (reset(pwNums[a], pw, charSetPoses, curSubKeySpace))
		{
			break;
		}
		for (b = 0, c = 0; b <= pwBlocks; b++, c += vecSize)
		{
			pwOut[c].data[a] = pw[b];
		}
	}
	return a;
}

bool KeySpace::cycleLastChar()
{
	uint32_t *charSetPoses, *pw;
	uint32_t offset, curSubKeySpace;
	int a, passwordLength;
	bool nextKeySpace = true;

	pw = m_pw + 1;
	curSubKeySpace = m_curSubKeySpace;
	passwordLength = (int) m_passwordLengths[curSubKeySpace];
	offset = m_offset;
	if (passwordLength > 1)
	{
		charSetPoses = m_charSetPoses;
		if ((m_charaterEncoding & CE_ASCII_BIN8) != 0)
		{
			// 8 bit
			for (a = passwordLength - 2; a >= 0; a--)
			{
				if (++charSetPoses[a] < m_charSetLengths[curSubKeySpace][a])
				{
					((char*)pw)[a] = m_subKeySpaces[curSubKeySpace][a][charSetPoses[a]];
					break;
				}
				charSetPoses[a] = 0;
				((char*)pw)[a] = m_subKeySpaces[curSubKeySpace][a][0];
			}
		}
		else if ((m_charaterEncoding & (CE_UTF16_LE | CE_UTF16_BE)) != 0)
		{
			// 16 bit
			for (a = passwordLength - 2; a >= 0; a--)
			{
				if (++charSetPoses[a] < m_charSetLengths[curSubKeySpace][a])
				{
					((char*)pw)[2 * a + offset] = m_subKeySpaces[curSubKeySpace][a][charSetPoses[a]];
					break;
				}
				charSetPoses[a] = 0;
				((char*)pw)[2 * a + offset] = m_subKeySpaces[curSubKeySpace][a][0];
			}
		}
		else // if ((m_charaterEncoding & (CE_UTF32_LE | CE_UTF32_BE)) != 0)
		{
			// 32 bit
			for (a = passwordLength - 2; a >= 0; a--)
			{
				if (++charSetPoses[a] < m_charSetLengths[curSubKeySpace][a])
				{
					((char*)pw)[4 * a + offset] = m_subKeySpaces[curSubKeySpace][a][charSetPoses[a]];
					break;
				}
				charSetPoses[a] = 0;
				((char*)pw)[4 * a + offset] = m_subKeySpaces[curSubKeySpace][a][0];
			}
		}
		nextKeySpace = a == -1;
	}
	if (nextKeySpace)
	{
		if (++curSubKeySpace >= m_numSubKeySpaces)
		{
			return true;
		}
		m_curSubKeySpace = curSubKeySpace;
		passwordLength = m_passwordLengths[curSubKeySpace];
		charSetPoses = m_charSetPoses;
		m_pw[0] = passwordLength;

		int pwBlocks = (int) m_pwBlocks;
		for (a = passwordLength >> 2; a < pwBlocks; a++)
		{
			pw[a] = 0;
		}

		if ((m_charaterEncoding & CE_ASCII_BIN8) != 0)
		{
			// 8 bit
			for (a = 0; a < passwordLength; a++)
			{
				charSetPoses[a] = 0;
				((char*)pw)[a] = m_subKeySpaces[curSubKeySpace][a][0];
			}
		}
		else if ((m_charaterEncoding & (CE_UTF16_LE | CE_UTF16_BE)) != 0)
		{
			// 16 bit
			for (a = 0; a < passwordLength; a++)
			{
				charSetPoses[a] = 0;
				((char*)pw)[2 * a + offset] = m_subKeySpaces[curSubKeySpace][a][0];
			}
		}
		else // if ((m_charaterEncoding & (CE_UTF32_LE | CE_UTF32_BE)) != 0)
		{
			// 32 bit
			for (a = 0; a < passwordLength; a++)
			{
				charSetPoses[a] = 0;
				((char*)pw)[4 * a + offset] = m_subKeySpaces[curSubKeySpace][a][0];
			}
		}
		if ((m_charaterEncoding & CE_80_NULL_PADDED) != 0)
		{
			if (m_charaterEncoding == (CE_ASCII_BIN8 | CE_80_NULL_PADDED))
			{
				((char*)pw)[a] = (char) 0x80;
			}
			else if ((m_charaterEncoding & (CE_UTF16_LE | CE_UTF16_BE)) != 0)
			{
				((char*)pw)[2 * a + offset] = (char) 0x80;
			}
			else // if ((m_charaterEncoding & (CE_UTF32_LE | CE_UTF32_BE)) != 0)
			{
				((char*)pw)[4 * a + offset] = (char) 0x80;
			}
		}
	}
	return false;
}

const uint64_t KeySpace::getKeySpace()
{
	return m_keySpace;
}

const uint32_t KeySpace::getMaxPwLen()
{
	return m_maxPwLen;
}

const uint32_t KeySpace::getMinPwLen()
{
	return m_minPwLen;
}
